import datetime
from functools import singledispatch
from itertools import dropwhile, filterfalse
from typing import Generator

from .date_range_utils import date_range
from .default_holiday_utils import global_default_holiday_discriminator
from .holiday_utils import IsHolidayFuncType


@singledispatch
def is_bizday(
    date: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> bool:
    """Check if the given date is a business day.

    Args:
        date (datetime.date): date to check.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Returns:
        bool: True if the date is a business day, False otherwise.
    """  # noqa: E501
    return not is_holiday(date)


@singledispatch
def get_next_bizday(
    date: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> datetime.date:
    """Get the next business day after the given date.

    Args:
        date (datetime.date): Reference date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Returns:
        datetime.date: Next business day after the given date.
    """  # noqa: E501
    return next(dropwhile(is_holiday, date_range(date, include_start=False)))


@singledispatch
def get_prev_bizday(
    date: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> datetime.date:
    """Get the previous business day before the given date.

    Args:
        date (datetime.date): Reference date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Returns:
        datetime.date: Previous business day before the given date.
    """  # noqa: E501
    return next(dropwhile(is_holiday, date_range(date, include_start=False, step_days=-1)))  # noqa: E501


@singledispatch
def get_n_next_bizday(
    date: datetime.date,
    n: int,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> datetime.date:
    """Get the n-th next business day after the given date.

    Args:
        date (datetime.date): Reference date.
        n (int): Number of business days to skip.
            0 means the same date.
            1 means the next business day.
            -1 means the previous business day.
            n > 1 means the n-th next business day.
            n < -1 means the (-n)-th previous business day.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Raises:
        ValueError: If n=0 and the date is a holiday.

    Returns:
        datetime.date: n-th next business day after the given date.

    Notes:
        - If n is negative, it will return the (-n)-th previous business day.
    """  # noqa: E501
    if n == 0:
        if is_holiday(date):
            raise ValueError(f"n=0 but date={date} is holiday")
        return date
    elif n > 0:
        return get_n_next_bizday(get_next_bizday(date, is_holiday), n - 1, is_holiday)  # noqa: E501
    else:
        return get_n_prev_bizday(date, -n, is_holiday)  # noqa: E501


@singledispatch
def get_n_prev_bizday(
    date: datetime.date,
    n: int,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> datetime.date:
    """Get the n-th previous business day before the given date."

    Args:
        date (datetime.date): Reference date.
        n (int): Number of business days to skip.
            0 means the same date.
            1 means the previous business day.
            -1 means the next business day.
            n > 1 means the n-th previous business day.
            n < -1 means the (-n)-th next business day.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Raises:
        ValueError: If n=0 and the date is a holiday.

    Returns:
        datetime.date: n-th previous business day before the given date.

    Notes:
        - If n is negative, it will return the (-n)-th next business day.
    """  # noqa: E501
    if n == 0:
        if is_holiday(date):
            raise ValueError(f"n=0 but date={date} is holiday")
        return date
    elif n > 0:
        return get_n_prev_bizday(get_prev_bizday(date, is_holiday), n - 1, is_holiday)  # noqa: E501
    else:
        return get_n_next_bizday(date, -n, is_holiday)  # noqa: E501


@singledispatch
def bizday_range(
    start: datetime.date,
    end: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    include_start: bool = True,
    include_end: bool = True,
) -> Generator[datetime.date, None, None]:
    """Generate a range of business days between two dates."

    Args:
        start (datetime.date): Start date.
        end (datetime.date): End date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.
        include_start (bool, optional): Include the start date in the range. Defaults to True.
        include_end (bool, optional): Include the end date in the range. Defaults to True.

    Yields:
        Generator[datetime.date, None, None]: Business days between start and end dates.

    Notes:
        - if include_start is True and start is not a holiday, the start date will be included in the range.
        - if include_end is True and end is not a holiday, the end date will be included in the range.
    """  # noqa: E501
    yield from filterfalse(
        is_holiday,
        date_range(
            start,
            end,
            include_start=include_start,
            include_end=include_end,
            step_days=1 if start <= end else -1,
        ),
    )


@singledispatch
def count_bizdays(
    start: datetime.date,
    end: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    include_start: bool = True,
    include_end: bool = True,
) -> int:
    """Count the number of business days between two dates.

    Args:
        start (datetime.date): Start date.
        end (datetime.date): End date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.
        include_start (bool, optional): Include the start date in the count. Defaults to True.
        include_end (bool, optional): Include the end date in the count. Defaults to True.

    Returns:
        int: Number of business days between start and end dates.

    Notes:
        - if start > end, the count will be negative.
        - if include_start is True and start is not a holiday, the start date will be included in the count.
        - if include_end is True and end is not a holiday, the end date will be included in the count.
    """  # noqa: E501
    if start > end:
        return - count_bizdays(end, start, is_holiday, include_end=include_start, include_start=include_end)  # noqa: E501
    bdrange = bizday_range(start, end, is_holiday, include_start=include_start, include_end=include_end)  # noqa: E501
    return sum(1 for _ in bdrange)



import datetime
from itertools import count
from typing import Callable, Generator


def date_range(
    start: datetime.date,
    end: datetime.date | None = None,
    *,
    include_start: bool = True,
    include_end: bool = True,
    step_days: int = 1,
) -> Generator[datetime.date, None, None]:
    """date generator from start to end.

    Args:
        start (datetime.date): start date
        end (datetime.date | None, optional): end date. Defaults to None.
        include_start (bool, optional): include start date. Defaults to True.
        include_end (bool, optional): include end date. Defaults to True.
        step_days (int, optional): step days. Defaults to 1

    Yields:
        Generator[datetime.date, None, None]: date generator

    Raises:
        ValueError: step_days is 0
    """
    # validate step_days
    if step_days == 0:
        raise ValueError("step_days must not be 0")

    # set ascending and delta
    ASCENDING = step_days > 0
    DELTA = datetime.timedelta(days=step_days)

    # set start date and end date
    if not include_start:
        start += DELTA
    if ASCENDING and end is None:
        end = datetime.date.max
    elif not ASCENDING and end is None:
        end = datetime.date.min
    assert end is not None

    # set end date
    is_broken: Callable[[datetime.date], bool] = {
        (True, True): end.__lt__,
        (True, False): end.__le__,
        (False, True): end.__gt__,
        (False, False): end.__ge__,
    }[(ASCENDING, include_end)]

    # yield date
    for _ in count():
        if is_broken(start):
            break
        yield start
        try:
            start += DELTA
        except OverflowError:
            break



import datetime
from typing import Callable

import pytest

from pybizday_utils.basic import (
    bizday_range,
    count_bizdays,
    get_n_next_bizday,
    get_n_prev_bizday,
    get_next_bizday,
    get_prev_bizday,
    is_bizday,
)


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, is_holiday, expected",
    [
        (datetime.date(2021, 1, 1), lambda d: d == datetime.date(2021, 1, 1), False),  # noqa
        (datetime.date(2021, 1, 2), lambda d: d == datetime.date(2021, 1, 3), True),  # noqa
    ],
)
def test_is_bizday(
    date: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    expected: bool,
) -> None:
    assert is_bizday(date, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            lambda _: False,
            datetime.date(2025, 4, 26),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 28),
        ),
    ],
)
def test_get_next_bizday(
    date: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_next_bizday(date, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            lambda _: False,
            datetime.date(2025, 4, 24),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d == (datetime.date(2025, 4, 24)),
            datetime.date(2025, 4, 23),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 24), datetime.date(2025, 4, 23)},  # noqa
            datetime.date(2025, 4, 22),
        ),
    ],
)
def test_get_prev_bizday(
    date: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_prev_bizday(date, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, n, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            1,
            lambda _: False,
            datetime.date(2025, 4, 26),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda _: False,
            datetime.date(2025, 4, 24),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d == (datetime.date(2025, 4, 24)),
            datetime.date(2025, 4, 23),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d in {datetime.date(2025, 4, 24), datetime.date(2025, 4, 23)},  # noqa
            datetime.date(2025, 4, 22),
        ),
        (
            datetime.date(2025, 4, 25),
            0,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda _: False,
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 29),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            datetime.date(2025, 4, 30),
        ),
        (
            datetime.date(2025, 4, 27),
            -2,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 28),
            -2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 29),
            -2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 30),
            -2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            datetime.date(2025, 4, 25),
        ),
    ],
)
def test_get_n_next_bizday(
    date: datetime.date,
    n: int,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_n_next_bizday(date, n, is_holiday) == expected


@pytest.mark.negative
def test_get_n_next_bizday_with_n_0_and_date_holiday() -> None:
    date = datetime.date(2025, 4, 25)
    def is_holiday(d: datetime.date) -> bool: return True
    with pytest.raises(ValueError):
        get_n_next_bizday(date, 0, is_holiday)


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, n, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda _: False,
            datetime.date(2025, 4, 26),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda _: False,
            datetime.date(2025, 4, 24),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d == (datetime.date(2025, 4, 24)),
            datetime.date(2025, 4, 23),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d in {datetime.date(2025, 4, 24), datetime.date(2025, 4, 23)},  # noqa
            datetime.date(2025, 4, 22),
        ),
        (
            datetime.date(2025, 4, 25),
            0,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda _: False,
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 29),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            datetime.date(2025, 4, 30),
        ),
        (
            datetime.date(2025, 4, 27),
            2,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 28),
            2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 29),
            2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 30),
            2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            datetime.date(2025, 4, 25),
        ),
    ],
)
def test_get_n_prev_bizday(
    date: datetime.date,
    n: int,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_n_prev_bizday(date, n, is_holiday) == expected


@pytest.mark.negative
def test_get_n_prev_bizday_with_n_0_and_date_holiday() -> None:
    date = datetime.date(2025, 4, 25)
    def is_holiday(d: datetime.date) -> bool: return True
    with pytest.raises(ValueError):
        get_n_prev_bizday(date, 0, is_holiday)


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, is_holiday, include_start, include_end, expected",
    [
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda _: False,
            True,
            True,
            [
                datetime.date(2025, 4, 25),
                datetime.date(2025, 4, 26),
                datetime.date(2025, 4, 27),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 29),
                datetime.date(2025, 4, 30),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            True,
            [
                datetime.date(2025, 4, 25),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 30),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            True,
            [
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 30),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            False,
            [
                datetime.date(2025, 4, 25),
                datetime.date(2025, 4, 28),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            False,
            [
                datetime.date(2025, 4, 28),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda _: False,
            True,
            True,
            [
                datetime.date(2025, 4, 30),
                datetime.date(2025, 4, 29),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 27),
                datetime.date(2025, 4, 26),
                datetime.date(2025, 4, 25),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            True,
            [
                datetime.date(2025, 4, 30),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 25),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            True,
            [
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 25),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            False,
            [
                datetime.date(2025, 4, 30),
                datetime.date(2025, 4, 28),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            False,
            [
                datetime.date(2025, 4, 28),
            ],
        )
    ],
)
def test_bizday_range(
    start: datetime.date,
    end: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    include_start: bool,
    include_end: bool,
    expected: list[datetime.date],
) -> None:
    actual = list(bizday_range(
        start,
        end,
        is_holiday,
        include_start=include_start,
        include_end=include_end,
    ))
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, is_holiday, include_start, include_end, expected",
    [
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda _: False,
            True,
            True,
            6,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            True,
            3,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            True,
            2,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            False,
            2,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            False,
            1,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda _: False,
            True,
            True,
            -6,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            True,
            -3,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            True,
            -2,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            True,
            False,
            -2,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27), datetime.date(2025, 4, 29)},  # noqa
            False,
            False,
            -1,
        )
    ],
)
def test_count_bizdays(
    start: datetime.date,
    end: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    include_start: bool,
    include_end: bool,
    expected: int,
) -> None:
    actual = count_bizdays(
        start,
        end,
        is_holiday,
        include_start=include_start,
        include_end=include_end,
    )
    assert actual == expected



import datetime

from dateutil.relativedelta import relativedelta

from .basic import (
    get_next_bizday,
    get_prev_bizday,
)
from .default_holiday_utils import global_default_holiday_discriminator
from .holiday_utils import IsHolidayFuncType


def is_biz_end_of_month(
    date: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> bool:
    """Check if the given date is the last business day of the month.

    Args:
        date (datetime.date): Date to check.
        is_holiday (IsHolidayFuncType, optional): Function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Returns:
        bool: True if the date is the last business day of the month, False otherwise.
    """  # noqa: E501
    if is_holiday(date):
        return False
    return date.month != (get_next_bizday(date, is_holiday)).month


def is_biz_start_of_month(
    date: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> bool:
    """Check if the given date is the first business day of the month.

    Args:
        date (datetime.date): Date to check.
        is_holiday (IsHolidayFuncType, optional): Function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Returns:
        bool: True if the date is the first business day of the month, False otherwise.
    """  # noqa: E501
    if is_holiday(date):
        return False
    return date.month != (get_prev_bizday(date, is_holiday)).month


def get_biz_end_of_month(
    date: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> datetime.date:
    """Get the last business day of the month for the given date.

    Args:
        date (datetime.date): reference date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Returns:
        datetime.date: last business day of the month for the given date.
    """  # noqa: E501
    date = date.replace(day=1)  # start of month
    date = date + datetime.timedelta(days=31)
    date = date.replace(day=1)  # start of next month
    return get_prev_bizday(date, is_holiday)


def get_biz_start_of_month(
    date: datetime.date,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
) -> datetime.date:
    """Get the first business day of the month for the given date.

    Args:
        date (datetime.date): reference date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.

    Returns:
        datetime.date: first business day of the month for the given date.
    """  # noqa: E501
    date = date.replace(day=1)  # start of month
    date = date - datetime.timedelta(days=1)  # end of previous month
    return get_next_bizday(date, is_holiday)


def add_years_months(
    date: datetime.date,
    years: int,
    months: int,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    bizeom2bizeom: bool = True,
    bizsom2bizsom: bool = False,
) -> datetime.date:
    """Add years and months to a date with business day adjustment.

    Args:
        date (datetime.date): Reference date.
        years (int): Years to add
        months (int): Months to add
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.
        bizeom2bizeom (bool, optional): Whether to adjust to the last business day of the month
            if the date is the last business day of the month. Defaults to True.
        bizsom2bizsom (bool, optional): Whether to adjust to the first business day of the month
            if the date is the first business day of the month. Defaults to False.

    Returns:
        datetime.date: Date after adding years and months with business day adjustment.
    """  # noqa: E501
    added = date + relativedelta(years=years, months=months)
    if bizeom2bizeom and is_biz_end_of_month(date, is_holiday):
        return get_biz_end_of_month(added, is_holiday)
    elif bizsom2bizsom and is_biz_start_of_month(date, is_holiday):
        return get_biz_start_of_month(added, is_holiday)
    else:
        return added


def add_years(
    date: datetime.date,
    years: int,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    bizeom2bizeom: bool = True,
    bizsom2bizsom: bool = False,
) -> datetime.date:
    """Add years to a date with business day adjustment.
    This function is a wrapper around add_years_months with months set to 0.

    Args:
        date (datetime.date): Reference date.
        years (int): Years to add
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.
        bizeom2bizeom (bool, optional): Whether to adjust to the last business day of the month
            if the date is the last business day of the month. Defaults to True.
        bizsom2bizsom (bool, optional): Whether to adjust to the first business day of the month
            if the date is the first business day of the month. Defaults to False.

    Returns:
        datetime.date: Date after adding years with business day adjustment.
    """  # noqa: E501
    return add_years_months(
        date,
        years,
        0,
        is_holiday,
        bizeom2bizeom=bizeom2bizeom,
        bizsom2bizsom=bizsom2bizsom,
    )


def add_months(
    date: datetime.date,
    months: int,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    bizeom2bizeom: bool = True,
    bizsom2bizsom: bool = False,
) -> datetime.date:
    """Add months to a date with business day adjustment.
    This function is a wrapper around add_years_months with years set to 0.

    Args:
        date (datetime.date): Reference date.
        months (int): Months to add
        is_holiday (IsHolidayFuncType, optional): function to check if a date is a holiday.
            Defaults to global_default_holiday_discriminator.
        bizeom2bizeom (bool, optional): Whether to adjust to the last business day of the month
            if the date is the last business day of the month. Defaults to True.
        bizsom2bizsom (bool, optional): Whether to adjust to the first business day of the month
            if the date is the first business day of the month. Defaults to False.

    Returns:
        datetime.date: Date after adding months with business day adjustment.
    """  # noqa: E501
    return add_years_months(
        date,
        0,
        months,
        is_holiday,
        bizeom2bizeom=bizeom2bizeom,
        bizsom2bizsom=bizsom2bizsom,
    )
